#include "CutGraph.h"

CutGraph::CutGraph(){;}
CutGraph::CutGraph(Mesh *mesh):cmesh(mesh){;}
CutGraph::~CutGraph(){;}

/* Compute the Dual Graph for the mesh
   The complexity is O(m), where m is the # of 
     edges in the mesh
*/
void CutGraph::ComputeDualGraph(void)
{
	//collect vertices for the dual graph
	int count = 0; // be the vid of the dual graph as well
	MeshFaceIterator mfitr(cmesh);
	for(;!mfitr.end();++mfitr){
		Face *f = *mfitr;
		fid2vid_dual[f->id()] = count;
		vid_dual2fid[count] = f->id();
		++count;
	}

	for(int i=0;i<count;i++){
		vector<int> v;
		AdjList_dual.push_back(v);
	}

	//collect adjacency list for the dual graph
	MeshEdgeIterator meitr(cmesh);
	for(;!meitr.end();++meitr){
		Edge *e = *meitr;
		int fid0 = e->halfedge(0)->face()->id();
		int fid1 = e->halfedge(1)->face()->id();
		// vid on the dual graph
		int vid0 = fid2vid_dual[fid0];
		int vid1 = fid2vid_dual[fid1];
		// edge on the dual graph
		//establish the mapping from edge on the mesh to the vid on the dual graph
		edgelist_dual.push_back(make_pair(vid0, vid1));
		//establish the connections for the dual graph
		AdjList_dual[vid0].push_back(vid1);
		AdjList_dual[vid1].push_back(vid0);
	}
}

/* Compute the Spananing Tree on the dual graph
   The complexity is O(n+m). Actually, it is O(c*(n+m)), since it is based on BFS,
     the complexity is O(n+m) for BFS.And c is the average # of neighbors for one vertex.
     In every iteration, it visits the neighbors of one vertex. But c is a constant for the mesh.
*/
void CutGraph::ComputeSpanningTree(void)
{
	//BFS tree
	vector<bool> visited_vertex; // vector for visited vertices

	int vsize = vid_dual2fid.size() ; 
	for(int i=0;i<vsize;i++){
		visited_vertex.push_back(false);
	}
	queue<int> q;
	q.push(0);
	visited_vertex[0] = true; //mark visited
	while (!q.empty()){
		int vid  = q.front();
		q.pop();
		for(int i=0;i<AdjList_dual[vid].size();i++){
			int vid_nb =AdjList_dual[vid][i];
			if (!visited_vertex[vid_nb]){
				q.push(vid_nb);
				//save the edge
				edgelist_span_tree.push_back(make_pair(vid, vid_nb));
				visited_vertex[vid_nb] = true; //mark the visted vertex
			}
		}
	}
	
	cout <<  "# of edges in spanning tree: " << edgelist_span_tree.size() << endl;
}

/* Compute the topological cut
   The complexity is O(m). Actually, it is O(c*m),
   where c is the # of neighbors of one vertex and it is usually 
   a constant around 5 or 6 in a mesh.
*/
void CutGraph::TopoCut()
{
	//initialize AdjList_dual_stree
	for(int i=0;i<fid2vid_dual.size();i++){
		vector<int> v;
		AdjList_dual_stree.push_back(v);
	}
	//initialize AdjList_topocut
	for(int i=0;i<cmesh->numVertices();i++){
		vector<int> v;
		AdjList_topocut.push_back(v);
	}

	for (int i=0;i<edgelist_span_tree.size();i++){
		pair<int, int> p = edgelist_span_tree[i];
		AdjList_dual_stree[p.first].push_back(p.second);
		AdjList_dual_stree[p.second].push_back(p.first);
	}

	//collect the edges whose dual are not in the the spanning tree
	//complexity: O(c*m), where m is the edges and c is the number 
	//of neighbors of one vertex, usually it is a constant around 5 or 6 in a mesh.
	MeshEdgeIterator meitr(cmesh);
	for(;!meitr.end();++meitr){
		Edge *e = *meitr;
		int fid0 = e->halfedge(0)->face()->id();
		int fid1 = e->halfedge(1)->face()->id();

		int vid0 = fid2vid_dual[fid0];
		int vid1 = fid2vid_dual[fid1];
		
		// try to find e=(fid0, fid1) -> (vid0, vid1) in the spanning tree
		bool found = false;
		vector<int> &v = AdjList_dual_stree[vid0];
		for(int i=0;i<v.size();i++){
			if (v[i] == vid1){ // the edges whose dual is in the spanning tree
				found = true;
				break;
			}
		}
		if(!found){
			int vid0 = e->vid(0);// vid on the mesh starts from 1 instead of 0
			int vid1 = e->vid(1);
			//establish the adjacency list for this topological cut
			AdjList_topocut[vid0-1].push_back(vid1-1);
			AdjList_topocut[vid1-1].push_back(vid0-1);
		}
		
	}
}


/* Trim Dangling Edges and output the result
	char* str: output fileanme for the cut graph edge list
  The complexity for Trimming is O(m). Actually it is O(d*2m), since it needs to iterate d times 
	to look up in the adjacency list, each time O(2m). The d is 
	the maximum depth of a "dangling path", for a specific graph,
	it is a constant.
  The complexity for output is O(mlogm), see the output part for details.
*/
void CutGraph::TrimDanglingEdge(char* str){
	int double_esize = 0; // size of the adjacency list
	for (int i=0;i<AdjList_topocut.size();i++){
		for (int j=0;j<AdjList_topocut[i].size();j++){
			double_esize++;
		}
	}
	
	//iterate to trim the dangling edges whose degree equals one, until the size of the adjacency list does not change
	int initial_count = double_esize+1;
	int itr_times = 0;
	for (int i=0; initial_count > double_esize;i++){
		initial_count = double_esize;
		++itr_times;
		for (int j=0;j<AdjList_topocut.size();j++){
			if(AdjList_topocut[j].size()==1){//dangling edge
				int vid = AdjList_topocut[j][0];
				AdjList_topocut[j].pop_back();
				//erase this edge need deal with two positions in the adj list
				int size = AdjList_topocut[vid].size();
				if(size == 1){
					AdjList_topocut[vid].pop_back();
					double_esize--;
				}else{
					for (int k=0;k<AdjList_topocut[vid].size();k++){
						if (AdjList_topocut[vid][k] == j){
							AdjList_topocut[vid][k] = AdjList_topocut[vid][size - 1]; //swap what we want to delete and the last entry
							break;
						}
					}
					AdjList_topocut[vid].pop_back();
					double_esize--;
				}
			}
		}
	}
	cout << "trimming the dangling edges iterates " << itr_times << " times!" << endl;

	//initalize
	for (int i=0;i<cmesh->numVertices();i++){
		vid4CutGraph.push_back(false);
	}


	//output edge list for the cut graph
	//the complexity is O(mlogm), since it iterates the adjacency list at a cost of O(m).
	//but during each iteration, it check the set and insert entry into the set, which cost O(logm)
	//However, if we use a boolean matrix to tag the edges we have processed, it will reduced to O(m).
	ofstream ofile(str);
	set<pair<int, int> > edge_set; // change it to a boolean matrix like // vector< vector<bool> > tag_matrix // initializing it cost O(m)
	for (int i=0;i<AdjList_topocut.size();i++){
		for(int j=0;j<AdjList_topocut[i].size();j++){
			pair<int, int> p0 = make_pair(i, AdjList_topocut[i][j]);
			pair<int, int> p1 = make_pair(AdjList_topocut[i][j], i);
			if(edge_set.find(p0)==edge_set.end() ){                    // if(tag_matrix[i][j] == true); //at the beginning, it is true
				edge_set.insert(p0);                                   // tag_matrix[i][j] = false;
				edge_set.insert(p1);                                   // tag_matrix[j][i] = false;
				ofile << "Edge " << i+1 << " " << AdjList_topocut[i][j]+1 << " {sharp}" << endl;
				//mark the vids on the cut graph
				vid4CutGraph[i] = true;
				vid4CutGraph[AdjList_topocut[i][j]] = true;
				vidOnCutGraph.push_back(i);
				vidOnCutGraph.push_back(AdjList_topocut[i][j]);
			}
		}
	}
	ofile.close();

}
bool IsMarkedHalfEdge(HalfEdge *he, vector<HalfEdge *> &vhe)
{
	for (int i=0;i<vhe.size();i++){
		if(vhe[i] == he)//if this HE is on the sharp edge
			return true;
	}
	return false;
}

int CutGraph::FindWid(int old_vid, int fid)
{
	vector<int> &vwids = old_vid2wids[old_vid];
	for (int i=0;i<vwids.size();i++){
		int wid = vwids[i];
		vector<int> &vfids = wid2fids[wid];
		for (int j=0;j<vfids.size();j++){
			if (vfids[j] == fid){
				return wid;
			}
		}
	}
	return -1;
}

void CutGraph::ComputeFundamentalDomain(char *ofilename)
{
	//Vertex and Face list for the fundamental domain
	vector<Vertex *> vlist;
	vector<Face *> flist;

	//Initialize
	for (int i=0;i<cmesh->numVertices();i++){
	//	vector<HalfEdge *> vhe;
	//	wedges.push_back(vhe);
		vector<int> vec;
		old_vid2wids.push_back(vec);
	}

	int wid = 0;
	MeshVertexIterator mvitr(cmesh);
	for (;!mvitr.end();mvitr++){
		Vertex *v = *mvitr;
		if (!vid4CutGraph[v->id()-1]){// not on the cut graph, 1 vertex for 1 wedge
			vector<HalfEdge *> vhe; //= wedges[wsize]; //collect wedges for this vertex

			HalfEdge *he = v->halfedge();
			if (he->source()->id() != v->id()){ //if it is not out-HE, get its twin
				he = he->he_twin();
			}
			//collect out-HEs
			HalfEdge *he_start = he;
			vector<int> vfid;
			do{
				vhe.push_back(he);
				vfid.push_back(he->face()->id());
				he = he->he_prev()->he_twin();
			} while (he!=he_start);
			
			wedges.push_back(vhe);
			wid2fids.push_back(vfid);
			wid2old_vid.push_back(v->id()-1);
			old_vid2wids[v->id() - 1].push_back(wid);
			++wid;
		}else{
			vector<HalfEdge *> vhe; //outHE of sharp edges around vertex v
			for (int i=0;i<AdjList_topocut[v->id()-1].size(); ++i){
				int vid = AdjList_topocut[v->id()-1][i];
				HalfEdge *he = cmesh->idHalfedge(v->id(), vid+1); // out HEs
				vhe.push_back(he);
			}

			// the # of vhes is the number of wedges for this vertex
			for (int i=0;i<vhe.size();i++){
				HalfEdge *he = vhe[i];
				HalfEdge *he_start = he;
				vector<HalfEdge *> neighbor_he; //collect the outHEs for a wedge
				vector<int> vfid;
				do{
					neighbor_he.push_back(he);	
					vfid.push_back(he->face()->id());
					he = he->he_prev()->he_twin();
				}while (!IsMarkedHalfEdge(he, vhe));
				wedges.push_back(neighbor_he);
				wid2fids.push_back(vfid);
				old_vid2wids[v->id() - 1].push_back(wid);
				wid2old_vid.push_back(v->id()-1);			
				wid++;
			}
		}
	}
	//create vertex list and face list
	ofstream ofile(ofilename);
	for (int i=0;i<wedges.size();i++){
		Vertex *v = new Vertex;
		Vertex *ver = wedges[i][0]->source();
		v->point() = ver->point();
		v->id() = ver->id();
		vlist.push_back(v);
		wid2new_vid.push_back(i);
		
		ofile << "Vertex " << i+1 << " " << v->point()[0] << " " << v->point()[1]<< " " << v->point()[2] << endl;
	}

	MeshFaceIterator mfitr(cmesh);
	for (;!mfitr.end();mfitr++)	{
		Face *f = *mfitr;

		Vertex *v[3];
		v[0] = f->halfedge()->source();
		v[1] = f->halfedge()->he_next()->source();
		v[2] = f->halfedge()->he_prev()->source();

		int new_vid[3];
		for (int i=0;i<3;i++){
			Vertex *ver = v[i];
			wid = FindWid(ver->id()-1, f->id());
			new_vid[i] = wid2new_vid[wid];
		}

		ofile << "Face " << f->id() << " " << new_vid[0]+1 << " " << new_vid[1]+1 << " " << new_vid[2]+1 << endl;
	}
	ofile.close();
	//clear
	for (int i=0;i<vlist.size();i++){
		delete vlist[i];
	}
}