#ifndef _XMESHLIB_BOUNDINGBOX_H_
#define _XMESHLIB_BOUNDINGBOX_H_

#include "../Math/Vector3.h"

namespace XMeshLib
{

	template<class _T>
	class Boundingbox
	{
	public:
		typedef Vector3<_T> Vector3t;

		Vector3t corner1;
		Vector3t corner2;

	public:
		Boundingbox() {}
		Boundingbox(const Vector3t &corner1, const Vector3t &corner2) 
		{ this->corner1 = corner1; this->corner2 = corner2; }
		Boundingbox(_T minX, _T maxX, _T minY, _T maxY, _T minZ, _T maxZ) 
		{ corner1.X(minX); corner2.X(maxX); corner1.Y(minY); corner2.Y(maxY); corner1.Z(minZ); corner2.Z(maxZ); }
		Boundingbox(const Vector3t &center, const _T dim)
		{ corner1 = Vector3t(center.X()-dim/2, center.Y()-dim/2, center.Z()-dim/2); corner2 = Vector3t(center.X()+dim/2, center.Y()+dim/2, center.Z()+dim/2); }
		Boundingbox(const Boundingbox &bbox) { corner1 = bbox.corner1; corner2 = bbox.corner2; }
		~Boundingbox() {}

		Vector3t GetCorner1() const { return corner1; }
		Vector3t GetCorner2() const { return corner2; }
		Vector3t center() const { return (corner1 + corner2)/2; }
		_T minX() const { return corner1.x; }
		_T maxX() const { return corner2.x; }
		_T minY() const { return corner1.y; }
		_T maxY() const { return corner2.y; }
		_T minZ() const { return corner1.z; }
		_T maxZ() const { return corner2.z; }
		_T & minX() { return corner1[0]; }
		_T & maxX() { return corner2[0]; }
		_T & minY() { return corner1[1]; }
		_T & maxY() { return corner2[1]; }
		_T & minZ() { return corner1[2]; }
		_T & maxZ() { return corner2[2]; }

		_T width() const { return corner2.x - corner1.x; }
		_T height() const { return corner2.y - corner1.y; }
		_T depth() const { return corner2.z - corner1.z; }

		_T diag() const { return (corner2 - corner1).norm(); }

		// test whether the bounding box contains the point
		bool IsContainingPoint(const Vector3t &point) const;
		// test whether the two bounding boxes intersect
		bool IsTouchingBox (const Boundingbox &box) const;
		// test whether the bounding box is completely outside another bounding box
		bool IsOutsideBox (const Boundingbox &box) const;
		// test whether the boudning box is completely inside another bounding box
		bool IsInsideBox (const Boundingbox &box) const;
		// test whether the another bounding box is completely contained in this bounding box
		bool IsContainingBox(const Boundingbox &box) const;
		// test whether the bounding box is empty
		bool IsEmptyBox() const;

		bool operator == (const Boundingbox &box) const 
		{ return (corner1 == box.corner1) && (corner2 == box.corner2); }


		// auto-coersion must exist from T2 to T
		// convert a bounding box of type T2 to a bounding box of type T
	/*	template<class T2>
		Boundingbox(const Boundingbox<T2>& box) { corner1 = box.corner1; corner2 = box.corner2; }
		template<class T2>
		Boundingbox& operator = (const Boundingbox<T2>& box) { corner1 = box.corner1; corner2 = box.corner2; return *this; }
	*/
		// enlarge the bounding box to contain the point as necessary
		bool contain(const Vector3t &point);

	};



	//////////////////////////////////////////////////////////////////////
	//		template implementation
	//////////////////////////////////////////////////////////////////////

	template<class _T>
	bool Boundingbox<_T>::IsContainingPoint(const Vector3t& point) const
	{
		return	(point.X() >= minX() && point.X() <= maxX() &&
				 point.Y() >= minY() && point.Y() <= maxY() &&
				 point.Z() >= minZ() && point.Z() <= maxZ());
	}

	template<class _T>
	bool Boundingbox<_T>::IsContainingBox(const Boundingbox<_T>& box) const
	{
		return	(box.minX() >= minX() && box.maxX() <= maxX() &&
				 box.minY() >= minY() && box.maxY() <= maxY() &&
				 box.minZ() >= minZ() && box.maxZ() <= maxZ());
	}

	template<class _T>
	bool Boundingbox<_T>::IsInsideBox(const Boundingbox<_T>& box) const
	{
		return	(minX() >= box.minX() && maxX() <= box.maxX() &&
				 minY() >= box.minY() && maxY() <= box.maxY() &&
				 minZ() >= box.minZ() && maxZ() <= box.maxZ());
	}

	template<class _T>
	bool Boundingbox<_T>::IsOutsideBox(const Boundingbox<_T>& box) const
	{
		return	(minX() >= box.maxX() || maxX() <= box.minX() ||
				 minY() >= box.maxY() || maxY() <= box.minY() ||
				 minZ() >= box.maxZ() || maxZ() <= box.minZ());
	}


	template<class _T>
	bool Boundingbox<_T>::IsTouchingBox(const Boundingbox<_T>& box) const
	{
		return (!IsContainingBox(box) && !IsInsideBox(box) && !IsOutsideBox(box));
	}


	template<class _T>
	bool Boundingbox<_T>::contain(const Vector3<_T> &point)
	{
		bool in = true;
		if (point.X() < corner1.X()) 
		{
			corner1[0] = point.X(); in = false;
		}
		else if (point.X() > corner2.X()) 
		{
			corner2[0] = point.X(); in = false; 
		}
		if (point.Y() < corner1.Y()) 
		{
			corner1[1] = point.Y(); in = false;
		}
		else if (point.Y() > corner2.Y()) 
		{
			corner2[1] = point.Y(); in = false;
		}
		if (point.Z() < corner1.Z()) 
		{
			corner1[2] = point.Z(); in = false; 
		}
		else if (point.Z() > corner2.Z()) 
		{
			corner2[2] = point.Z(); in = false;
		}
		return in;
	}



	template<class _T>
	bool Boundingbox<_T>::IsEmptyBox() const
	{
		return (corner1 == corner2);
	}



	//-------------------------------------------

	typedef Boundingbox<int> Boundingboxi;
	typedef Boundingbox<float> Boundingboxf;
	typedef Boundingbox<double> Boundingboxd;

}

#endif //_XMESHLIB_BOUNDINGBOX_H_
