#include "DisplayWindow.h"
#include "gui.h"
#include "../Core_Mesh/Mesh.h"
#include "../Core_Mesh/Iterators.h"
#include "../MeshFuncLib/GetNormals.h"

extern Gui *gui;

using namespace XMeshLib;

// the constructor method: w and h are the window width and height, respectively.
DisplayWindow::DisplayWindow(int x,int y,int w,int h,const char *l) 
  : Fl_Gl_Window(x,y,w,h,l)
{
   // clear window 
   glClearColor(0.0,0.0,0.0,0.0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   
   glInitialized = false;
   theMesh = NULL;
   InitSwitches();
}


//Initialization
void DisplayWindow::InitSwitches()
{
	glInitialized = false;

	rotate_dist = 0;
	view_dist = 0;

	shading_mode = SHADING_SMOOTH;
	material_mode = MATERIAL_DEFAULT;
	polygon_mode = POLYGON_FILL;

	enable_lighting = 1;
	enable_picking = 0;
	picking_mode = PICKING_VERTEX;

	show_normal = 0;
	show_boundingbox = 0;
	show_triangulation = 0;
	show_sharpvert = 1;
	show_sharpedge = 1;
	show_sharpface = 1;
	show_texcoord = 0;	
	show_envmap = 0;
	show_floor = 0;
	show_stats = 1;

	gl_version = NULL;
	gl_vendor = NULL;
	gl_renderer = NULL;
	gl_extensions = NULL;
	glu_extensions = NULL;

	win_width = 0;
	win_height = 0;
	fullscreen = false;

	//R = 0.6f; G= 0.6f; B = 0.6f;

	for (int i = 0; i < 256; ++i)
		key_down[i] = 0;
}

void DisplayWindow::InitGL()
{
	gl_version = glGetString(GL_VERSION);
	gl_vendor = glGetString(GL_VENDOR);
	gl_renderer = glGetString(GL_RENDERER);
	gl_extensions = glGetString(GL_EXTENSIONS);
	glu_extensions = gluGetString(GLU_EXTENSIONS);

	glClearColor(R, G, B, 0.0f);
	glPolygonMode(GL_FRONT, GL_FILL);
	glShadeModel(GL_SMOOTH);

	glEnable(GL_DEPTH_TEST);

	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);
	//glEnable(GL_CULL_FACE);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	// set pixel alignment
	glPixelStorei(GL_PACK_ALIGNMENT, 1);

	InitLighting();
	InitMaterial();
	InitQuadrics();
	InitTextureData();	
	InitCamera();
}

void DisplayWindow::InitTextureData()
{
	m_ctexture2D.LoadFile("checkerboard1024.bmp");
}

void DisplayWindow::InitMaterial(void)
{
	SetMaterial(MATERIAL_DEFAULT);
}

void DisplayWindow::InitQuadrics()
{
	quadric_list = glGenLists(4);
	
	GLUquadricObj *qobj;
	qobj = gluNewQuadric();
	gluQuadricCallback(qobj, GLU_ERROR, (GluQuadricCallbackType)quadric_error_cb);

	gluQuadricDrawStyle(qobj, GLU_FILL);
	gluQuadricNormals(qobj, GLU_SMOOTH);

	// A sphere quadric
	glNewList(quadric_list, GL_COMPILE);
		gluSphere(qobj, 0.02, 30, 20);
	glEndList();

	// A cone quadric
	glNewList(quadric_list+1, GL_COMPILE);
		gluCylinder(qobj, 0.04, 0, 0.08, 30, 20);
	glEndList();

	// A disk quadric
	glNewList(quadric_list+2, GL_COMPILE);
		gluDisk(qobj, 0, 0.04, 30, 20);
	glEndList();
	
	// A cylinder quadric
	glNewList(quadric_list+3, GL_COMPILE);
		gluCylinder(qobj, 0.02, 0.02, 0.4, 30, 20);
	glEndList();		
}





void DisplayWindow::SetMesh(Mesh * inmesh)
{
	theMesh = inmesh;
}

void DisplayWindow::draw() 
{
	// Fl_Gl_Window::valid() is turned off when FLTK creates a new context 
	// for this window or when the window resizes, and is turned on after draw() is called.
	if (!glInitialized)
	{
		InitGL();
	
		GLenum err = glewInit();
		if (GLEW_OK != err)
		{
			/* Problem: glewInit failed, something is seriously wrong. */
			fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
			return;
		}
		fprintf(stdout, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));
		glInitialized= true;
	}
	ProcessRendering();
}

void DisplayWindow::ProcessRendering()
{
	assert(glInitialized);
	Reshape(w(), h());

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	theCamera.SetGL(view_dist/DOF, view_dist*DOF, win_width, win_height);

	if (shading_mode == SHADING_SMOOTH)
		glShadeModel(GL_SMOOTH);
	else 
		glShadeModel(GL_FLAT);

	RenderMesh();
	if (show_sharpvert)
		RenderSharpVertex();
	if (show_sharpedge)
		RenderSharpEdge();
	if (show_triangulation)
		RenderTessellation();
	if (show_normal)
		RenderNormal();
	if (show_boundingbox)
		RenderBoundingBox();
	//if (show_floor)
		//	RenderFloor();
	glPopMatrix();
}

void DisplayWindow::RenderMesh()
{
	if (!theMesh) 
		return;
	int cur_mat = material_mode;

	//glEnable(GL_COLOR_MATERIAL);
	glDisable(GL_COLOR_MATERIAL);
	SetMaterial(cur_mat);

	if (show_envmap)
	{
		//BeginEnvMap();
		//glEnable(GL_TEXTURE_2D);
		//glBindTexture(GL_TEXTURE_2D, texture2_name[1]);
	}
	else if (show_texcoord)
	{		
		glDisable(GL_COLOR_MATERIAL);
		SetMaterial(cur_mat);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, m_ctexture2D.TexName());//texture_name[0]);
	}

	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glPolygonMode(GL_FRONT, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_LINE:
		glPolygonMode(GL_FRONT, GL_LINE);
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_POINT:
		glPolygonMode(GL_FRONT, GL_POINT);
		glEnable(GL_POLYGON_OFFSET_POINT);
		glPolygonOffset(1.0, 1.0);
		glPointSize(3.0);
		break;
	}

	if (!theMesh->hasNormal())
	{
		GetNormals::GetNorm(theMesh, true);
	}
	//glColor3f(crgb[0], crgb[1], crgb[2]);
	glBegin(GL_TRIANGLES);

	for(MeshFaceIterator fiter(theMesh); !fiter.end(); ++fiter)
	{
		Face *f = *fiter;
		HalfEdge * he = f->halfedge();
		Vertex * v[3];
		v[0] = he->source();
		v[1] = he->target();
		v[2] = he->he_next()->target();

		double val[2];

		for (int i=0;i<3;++i)
		{
			Point & pt = v[i]->point();
			Point & npt = theMesh->verNormal(v[i]);

			if (show_texcoord)
			{
				TraitParser::ReadTrait(v[i]->string(),"uv",2,val);
				glTexCoord2d (val[0],val[1]);
			}
			glNormal3d( npt[0], npt[1], npt[2]);
			glVertex3d(pt[0], pt[1], pt[2]);
		}
	}
	glEnd();
	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glDisable(GL_POLYGON_OFFSET_FILL);
		break;
	case POLYGON_LINE:
		glDisable(GL_POLYGON_OFFSET_LINE);
		break;
	case POLYGON_POINT:
		glDisable(GL_POLYGON_OFFSET_POINT);
		glPointSize(1.0);
		break;
	}
}

void DisplayWindow::RenderFloor()
{
	const double INFINITY = 100;
	double y;

	y = m_bbox.minY();

	DoDisableLights();

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture2_name[2]);

	glColor3f(1.0f, 1.0f, 1.0f);

	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3d(-INFINITY, y, INFINITY);
		glTexCoord2f(0.0f, 64.0f);	
		glVertex3d(INFINITY, y, INFINITY);
		glTexCoord2f(64.0f, 64.0f);
		glVertex3d(INFINITY, y, -INFINITY);
		glTexCoord2f(64.0f, 0.0f);
		glVertex3d(-INFINITY, y, -INFINITY);
	glEnd();


    glDisable(GL_TEXTURE_2D);
	if (enable_lighting)
	{
		DoEnableLights();
	}
}

void DisplayWindow::RenderTextOnPoint(const char text[], Point & pos)
{	
	// raster fonts
	DoDisableLights();

	glLineWidth(1.0);

	glColor3f(0.0f, 0.0f, 0.0f);
	glPushMatrix();
	glTranslatef(pos[0], pos[1], pos[2]);
	glScalef(0.1f, 0.1f, 0.1f);
	RenderText(text, GLUT_STROKE_MONO_ROMAN, FONT_STROKE);	
	glPopMatrix();
	glFlush();

	if (enable_lighting)
	{
		DoEnableLights();
	}
}


void DisplayWindow::RenderBall(int mat_mode, float size)
{
	int cur_mat = material_mode;
	SetMaterial(mat_mode);

	glShadeModel(GL_SMOOTH);
	
	glPushMatrix();
		glScalef(size, size, size);
		glCallList(quadric_list);
	glPopMatrix();

	if (shading_mode == SHADING_SMOOTH)
		glShadeModel(GL_SMOOTH);
	else 
		glShadeModel(GL_FLAT);

	SetMaterial(cur_mat);
}


void DisplayWindow::RenderBall(float R, float G, float B, float size)
{
	int cur_mat = material_mode;
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_SPECULAR);
	glColor3f(0.7f, 0.7f, 0.7f);
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(0.3f, 0.3f, 0.3f);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	glColor3f(R, G, B);

	glShadeModel(GL_SMOOTH);

	glPushMatrix();
		glScalef(size, size, size);
		glCallList(quadric_list);
	glPopMatrix();

	if (shading_mode == SHADING_SMOOTH)
		glShadeModel(GL_SMOOTH);
	else 
		glShadeModel(GL_FLAT);

	glDisable(GL_COLOR_MATERIAL);
	SetMaterial(cur_mat);
}


void DisplayWindow::RenderArrow(float R, float G, float B, float size)
{
	int cur_mat = material_mode;
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_SPECULAR);
	glColor3f(0.7f, 0.7f, 0.7f);
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(0.3f, 0.3f, 0.3f);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	glColor3f(R, G, B);
	
	glPushMatrix();
		glScalef(size, size, size);
		glPushMatrix();
		glTranslatef(0.0f,0.0f, 0.4f);
		glCallList(quadric_list+1);		
		glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
		glCallList(quadric_list+2);
		glPopMatrix();

		glCallList(quadric_list+3);
		glScalef(0.5f, 0.5f, 0.5f);
		glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
		glCallList(quadric_list+2);
	glPopMatrix();

	glDisable(GL_COLOR_MATERIAL);
	SetMaterial(cur_mat);
}


void DisplayWindow::RenderCoordinate()
{
	glPushMatrix();
		glPushMatrix();
		glRotated(90, 0, 1, 0);
		RenderArrow(0.8f, 0.0f, 0.0f);
		glPopMatrix();

		glPushMatrix();
		glRotated(-90, 1, 0, 0);
		RenderArrow(0.0f, 0.8f, 0.0f);
		glPopMatrix();

		RenderArrow(0.0f, 0.0f, 0.8f);


	glPopMatrix();
}

void DisplayWindow::RenderText(const char *text, void *font, int font_flag)
{
	if (!text)
		return;

	int nchar = strlen(text);

	int i;
	if (font_flag == FONT_STROKE)
	{
		for (i = 0; i < nchar; ++i)
			glutStrokeCharacter(font, text[i]);
	}
	else
	{
		for (i = 0; i < nchar; ++i)
			glutBitmapCharacter(font, text[i]);
	}
}

void DisplayWindow::RenderBoundingBox()
{
	Vector3d blb, trf, centroid;

	blb = m_bbox.GetCorner1();
	trf = m_bbox.GetCorner2();
	centroid = m_bbox.center();

	DoDisableLights();
	glColor3f(0.6f, 0.6f, 0.6f);

	glLineWidth(2.0);
	glBegin(GL_LINES);
		glVertex3d(blb.x, blb.y, blb.z);	glVertex3d(trf.x, blb.y, blb.z);
		glVertex3d(blb.x, blb.y, trf.z);	glVertex3d(trf.x, blb.y, trf.z);
		glVertex3d(blb.x, trf.y, trf.z);	glVertex3d(trf.x, trf.y, trf.z);
		glVertex3d(blb.x, trf.y, blb.z);	glVertex3d(trf.x, trf.y, blb.z);

		glVertex3d(blb.x, blb.y, blb.z);	glVertex3d(blb.x, blb.y, trf.z);
		glVertex3d(blb.x, trf.y, blb.z);	glVertex3d(blb.x, trf.y, trf.z);
		glVertex3d(trf.x, trf.y, blb.z);	glVertex3d(trf.x, trf.y, trf.z);
		glVertex3d(trf.x, blb.y, blb.z);	glVertex3d(trf.x, blb.y, trf.z);

		glVertex3d(blb.x, blb.y, blb.z);	glVertex3d(blb.x, trf.y, blb.z);
		glVertex3d(trf.x, blb.y, blb.z);	glVertex3d(trf.x, trf.y, blb.z);
		glVertex3d(trf.x, blb.y, trf.z);	glVertex3d(trf.x, trf.y, trf.z);
		glVertex3d(blb.x, blb.y, trf.z);	glVertex3d(blb.x, trf.y, trf.z);
	glEnd();
	glLineWidth(1.0);

	if (enable_lighting)
	{
		DoEnableLights();
	}

	glPushMatrix();
	glTranslated(blb.x, blb.y, blb.z);
	RenderCoordinate();
	glPopMatrix();

	glPushMatrix();
	glTranslated(centroid.x, centroid.y, centroid.z);
	RenderBall(1.0f, 1.0f, 1.0f, 2);
	glPopMatrix();

}
void DisplayWindow::RenderNormal()
{
	;
}
void DisplayWindow::RenderTessellation()
{
	if (!theMesh) return;

	// Draw the hidden lines (edges)
	DoDisableLights();
	glColor3f(0.0f, 0.0f, 0.0f);

	// anti-aliasing
	glEnable (GL_LINE_SMOOTH);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	glBegin(GL_LINES);		
	for(MeshEdgeIterator eiter(theMesh); !eiter.end(); ++eiter)
	{
		Edge * e = *eiter;
		Vertex * v1 = theMesh->edgeVertex1(e);
		Vertex * v2 = theMesh->edgeVertex2(e);
		Point & p1 = v1->point(); 
		Point & p2 = v2->point();

		if (e->boundary())
			glColor3f(1.0f, 0.0f, 0.0f);

		if (e->sharp())
			glColor3f(1.0f, 1.0f, 0.0f);

		glVertex3f(p1[0], p1[1], p1[2]);
		glVertex3f(p2[0], p2[1], p2[2]);

		glColor3f(0.0f, 0.0f, 0.0f);
	
	}
	glEnd();
	glColor3f(surfaceColor[0], surfaceColor[1], surfaceColor[2]);
	if (enable_lighting)
		this->DoEnableLights();

	glDisable (GL_LINE_SMOOTH);
	glDisable (GL_BLEND);
}

void DisplayWindow::RenderSharpVertex()
{
	if (!theMesh)
		return;
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	DoDisableLights();
	float radius = m_bbox.diag() / 200 * 50;

	if (radius > 0.7f) radius = 0.7f;
	if (radius < 0.3f) radius = 0.3f;

	for (int i=0;i<pickedVs.size(); ++i)
	{
		Vertex * ver = pickedVs[i];
		Point & p = ver->point();
		glPushMatrix();
			glTranslated(p[0], p[1], p[2]);
			RenderBall(sharpVerColor[0], sharpVerColor[1], sharpVerColor[2], radius);				
		glPopMatrix();
	}
	if (enable_lighting)
	{
		DoEnableLights();
	}
	glDisable(GL_COLOR_MATERIAL);
}


void DisplayWindow::RenderSharpEdge()
{
	if (!theMesh) 
		return;

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);	

	// Draw the hidden lines (edges)
	DoDisableLights();

	// anti-aliasing
	glEnable (GL_LINE_SMOOTH);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	glLineWidth(2.0);
	glColor3f(sharpEdgeColor[0], sharpEdgeColor[1], sharpEdgeColor[2]);

	glBegin(GL_LINES);		
	for(MeshEdgeIterator eiter(theMesh); !eiter.end(); ++eiter)
	{
		Edge * e = *eiter;
		if (!e->sharp())
			continue;

		Vertex * v1 = theMesh->edgeVertex1(e);
		Vertex * v2 = theMesh->edgeVertex2(e);
		Point & p1 = v1->point(); 
		Point & p2 = v2->point();

		glVertex3f(p1[0], p1[1], p1[2]);
		glVertex3f(p2[0], p2[1], p2[2]);	
	}
	glEnd();
	glLineWidth(1.0);

	if (enable_lighting)
	{
		DoEnableLights();
	}

	glDisable (GL_LINE_SMOOTH);
	glDisable (GL_BLEND);

	glDisable(GL_COLOR_MATERIAL);
}

void DisplayWindow::RenderSharpFace()
{
	if (!theMesh)
		return;

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	DoDisableLights();

	glBegin(GL_TRIANGLES);

	for (int i=0;i<pickedFs.size(); ++i)
	{
		Face * f = pickedFs[i];
		HalfEdge * he = f->halfedge();
		Point & p0 = he->source()->point();
		Point & p1 = he->target()->point();
		Point & p2 = he->he_next()->target()->point();
		glVertex3d(p0[0], p0[1], p0[2]);
		glVertex3d(p1[0], p1[1], p1[2]);
		glVertex3d(p2[0], p2[1], p2[2]);
	}
	glEnd();
	if (enable_lighting)
	{
		DoEnableLights();
	}
	glDisable(GL_COLOR_MATERIAL);
}


void DisplayWindow::RenderMesh4Picking(int pmode)
{
	if (!theMesh) 
		return;
	
	bool flag = false;

	MeshFaceIterator fiter(theMesh);
	MeshVertexIterator viter(theMesh);
	MeshEdgeIterator eiter(theMesh);

	switch (pmode)
	{
	case PICKING_FACE:
		for(; !fiter.end(); ++fiter)
		{
			Face *f = *fiter;
			glLoadName(f->id());
			glBegin(GL_TRIANGLES);
			for(FaceVertexIterator fviter(f); !fviter.end(); ++fviter)
			{
				Vertex *v = *fviter;
				Point & p = v->point();

				glVertex3d(p[0], p[1], p[2]);
			}
			glEnd();
		}
		break;
	case PICKING_VERTEX:
		for(; !viter.end(); ++viter)
		{
			Vertex *v = *viter;
			glLoadName(v->id());
			Point &p = v->point();			
			glBegin(GL_POINTS);
			glVertex3d( p[0], p[1], p[2] );
			glEnd();
		}
		break;
	case PICKING_EDGE:
		for(; !eiter.end(); ++eiter )
		{
			Edge * e = *eiter;
			Vertex * v1 = theMesh->edgeVertex1(e);
			Vertex * v2 = theMesh->edgeVertex2(e);

			int vid1 = v1->id(); //e->vid(0); //e->vertex(0);
			int vid2 = v2->id(); //e->vid(1); //vertex(1);
			// Note, the order of pushing names onto stack is important!
			if (!flag)
			{
				glPopName();
				glPushName(vid2);
				glPushName(vid1);
				flag = true;
			}
			else
			{
				glPopName(); 
				glPopName();
				glPushName(vid2);
				glPushName(vid1);
			}
			

			Point & p1 = v1->point();
			Point & p2 = v2->point();

			glBegin(GL_LINES);
			glVertex3f(p1[0], p1[1], p1[2]);
			glVertex3f(p2[0], p2[1], p2[2]);
			glEnd();
		}		
		break;
	default:
		break;
	}
}



void DisplayWindow::InitCamera()
{
	Vector3f homepos, homerotaxis;

	// Since the model is always normalized, constants suffice
	homepos.x = 0;
	homepos.y = 0;
	homepos.z = 3.0;
	
	homerotaxis = Vector3f(0, 0, 1);

	theCamera.SetHome(homepos, 0, homerotaxis, DEFAULT_FOV);
	rotate_dist = view_dist = 3.0;

	theCamera.GoHome();
}

void DisplayWindow::InitLighting()
{
	GLfloat	light_Ka[] = {0.3f,0.3f,0.3f,1.0f};
	GLfloat	light_Kd[] = {0.8f,0.8f,0.8f,1.0f};
	GLfloat light_Ks[]= {0.8f,0.8f,0.8f,1.0f};
	
	GLfloat light_Pos[4][4] = {0.0f,5.0f,5.0f,0.0f,
		0.0f,-5.0f,6.0f,0.0f,
		0.0f,0.0f,10.0f,0.0f,
		0.0f,0.0f,-10.0f,0.0f};

	glPushMatrix();

	glLightfv(GL_LIGHT0, GL_AMBIENT, light_Ka);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_Kd);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_Ks);
	glLightfv(GL_LIGHT0, GL_POSITION, light_Pos[0]);
	glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT1, GL_AMBIENT, light_Ka);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light_Kd);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light_Ks);
	glLightfv(GL_LIGHT1, GL_POSITION, light_Pos[1]);
	glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT2, GL_AMBIENT, light_Ka);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, light_Kd);
	glLightfv(GL_LIGHT2, GL_SPECULAR, light_Ks);
	glLightfv(GL_LIGHT2, GL_POSITION, light_Pos[2]);
	glLightf(GL_LIGHT2, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT2, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT2, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT3, GL_AMBIENT, light_Ka);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, light_Kd);
	glLightfv(GL_LIGHT3, GL_SPECULAR, light_Ks);
	glLightfv(GL_LIGHT3, GL_POSITION, light_Pos[3]);
	glLightf(GL_LIGHT3, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT3, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT3, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	// separate specular component
	glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
	glPopMatrix();

	lights_on[0] = true;
	lights_on[1] = false;
	lights_on[2] = false;
	lights_on[3] = false;

	if (enable_lighting)
	{
		DoEnableLights();
	}
	
	for (int i=0;i<4; ++i)
		for (int j=0;j<4;++j)
			light_pos[i][j] = light_Pos[i][j];
	return;
}


void DisplayWindow::SetMaterial(int material)
{
	GLfloat mat_Ka[4];
	GLfloat mat_Kd[4];
	GLfloat mat_Ks[4];
	GLfloat mat_Se;
	mat_Ka[3] = 1.0f;
	mat_Kd[3] = 1.0f;
	mat_Ks[3] = 1.0f;

	// Record material
	material_mode = material;

	switch(material)
	{
	case MATERIAL_BRASS:
		mat_Ka[0] = 0.329412f;
		mat_Kd[0] = 0.780392f;
		mat_Ks[0] = 0.992157f;
		mat_Ka[1] = 0.223529f;
		mat_Kd[1] = 0.568627f;
		mat_Ks[1] = 0.941176f;
		mat_Ka[2] = 0.027451f;
		mat_Kd[2] = 0.113725f;
		mat_Ks[2] = 0.807843f;
		mat_Se = 27.8974f;
		break;

	case MATERIAL_BRONZE:
		mat_Ka[0] = 0.2125f;
		mat_Kd[0] = 0.714f;
		mat_Ks[0] = 0.393548f;
		mat_Ka[1] = 0.1275f;
		mat_Kd[1] = 0.4284f;
		mat_Ks[1] = 0.271906f;
		mat_Ka[2] = 0.054f;
		mat_Kd[2] = 0.18144f;
		mat_Ks[2] = 0.166721f;
		mat_Se = 25.6f;
		break;

	case MATERIAL_CHROME:
		mat_Ka[0] = 0.25f;
		mat_Kd[0] = 0.4f;
		mat_Ks[0] = 0.774597f;
		mat_Ka[1] = 0.25f;
		mat_Kd[1] = 0.4f;
		mat_Ks[1] = 0.774597f;
		mat_Ka[2] = 0.25f;
		mat_Kd[2] = 0.4f;
		mat_Ks[2] = 0.774597f;
		mat_Se = 76.8f;
		break;

	case MATERIAL_COPPER:
		mat_Ka[0] = 0.19125f;
		mat_Kd[0] = 0.7038f;
		mat_Ks[0] = 0.256777f;
		mat_Ka[1] = 0.0735f;
		mat_Kd[1] = 0.27048f;
		mat_Ks[1] = 0.137622f;
		mat_Ka[2] = 0.0225f;
		mat_Kd[2] = 0.0828f;
		mat_Ks[2] = 0.086014f;
		mat_Se = 12.84f;
		break;

	case MATERIAL_GOLD:
		mat_Ka[0] = 0.24725f;
		mat_Kd[0] = 0.75164f;
		mat_Ks[0] = 0.628281f;
		mat_Ka[1] = 0.1995f;
		mat_Kd[1] = 0.60648f;
		mat_Ks[1] = 0.555802f;
		mat_Ka[2] = 0.0745f;
		mat_Kd[2] = 0.22648f;
		mat_Ks[2] = 0.366065f;
		mat_Se = 51.2f;
		break;

	case MATERIAL_P_GOLD:
		mat_Ka[0] = 0.24725f;
		mat_Kd[0] = 0.34615f;
		mat_Ks[0] = 0.797357f;
		mat_Ka[1] = 0.2245f;
		mat_Kd[1] = 0.3143f;
		mat_Ks[1] = 0.723991f;
		mat_Ka[2] = 0.0645f;
		mat_Kd[2] = 0.0903f;
		mat_Ks[2] = 0.208006f;
		mat_Se = 83.2f;
		break;

	case MATERIAL_PEWTER:
		mat_Ka[0] = 0.105882f;
		mat_Kd[0] = 0.427451f;
		mat_Ks[0] = 0.333333f;
		mat_Ka[1] = 0.058824f;
		mat_Kd[1] = 0.470588f;
		mat_Ks[1] = 0.333333f;
		mat_Ka[2] = 0.113725f;
		mat_Kd[2] = 0.541176f;
		mat_Ks[2] = 0.521569f;
		mat_Se = 9.84615f;
		break;

	case MATERIAL_SILVER:
		mat_Ka[0] = 0.19225f;
		mat_Kd[0] = 0.50754f;
		mat_Ks[0] = 0.508273f;
		mat_Ka[1] = 0.19225f;
		mat_Kd[1] = 0.50754f;
		mat_Ks[1] = 0.508273f;
		mat_Ka[2] = 0.19225f;
		mat_Kd[2] = 0.50754f;
		mat_Ks[2] = 0.508273f;
		mat_Se = 51.2f;
		break;

	case MATERIAL_P_SILVER:
		mat_Ka[0] = 0.23125f;
		mat_Kd[0] = 0.2775f;
		mat_Ks[0] = 0.773911f;
		mat_Ka[1] = 0.23125f;
		mat_Kd[1] = 0.2775f;
		mat_Ks[1] = 0.773911f;
		mat_Ka[2] = 0.23125f;
		mat_Kd[2] = 0.2775f;
		mat_Ks[2] = 0.773911f;
		mat_Se = 89.6f;
		break;

	case MATERIAL_EMERALD:
		mat_Ka[0] = 0.0215f;
		mat_Kd[0] = 0.07568f;
		mat_Ks[0] = 0.633f;
		mat_Ka[1] = 0.1745f;
		mat_Kd[1] = 0.61424f;
		mat_Ks[1] = 0.727811f;
		mat_Ka[2] = 0.0215f;
		mat_Kd[2] = 0.07568f;
		mat_Ks[2] = 0.633f;
		mat_Ka[3] = 0.55f;
		mat_Kd[3] = 0.55f;
		mat_Ks[3] = 0.55f;
		mat_Se = 76.8f;
		break;

	case MATERIAL_JADE:
		mat_Ka[0] = 0.135f;
		mat_Kd[0] = 0.54f;
		mat_Ks[0] = 0.316228f;
		mat_Ka[1] = 0.2225f;
		mat_Kd[1] = 0.89f;
		mat_Ks[1] = 0.316228f;
		mat_Ka[2] = 0.1575f;
		mat_Kd[2] = 0.63f;
		mat_Ks[2] = 0.316228f;
		mat_Ka[3] = 0.95f;
		mat_Kd[3] = 0.95f;
		mat_Ks[3] = 0.95f;
		mat_Se = 12.8f;
		break;

	case MATERIAL_OBSIDIAN:
		mat_Ka[0] = 0.05375f;
		mat_Kd[0] = 0.18275f;
		mat_Ks[0] = 0.332741f;
		mat_Ka[1] = 0.05f;
		mat_Kd[1] = 0.17f;
		mat_Ks[1] = 0.328634f;
		mat_Ka[2] = 0.06625f;
		mat_Kd[2] = 0.22525f;
		mat_Ks[2] = 0.346435f;
		mat_Ka[3] = 0.82f;
		mat_Kd[3] = 0.82f;
		mat_Ks[3] = 0.82f;
		mat_Se = 38.4f;
		break;

	case MATERIAL_PEARL:
		mat_Ka[0] = 0.25f;
		mat_Kd[0] = 1.0f;
		mat_Ks[0] = 0.296648f;
		mat_Ka[1] = 0.20725f;
		mat_Kd[1] = 0.829f;
		mat_Ks[1] = 0.296648f;
		mat_Ka[2] = 0.20725f;
		mat_Kd[2] = 0.829f;
		mat_Ks[2] = 0.296648f;
		mat_Ka[3] = 0.922f;
		mat_Kd[3] = 0.922f;
		mat_Ks[3] = 0.922f;
		mat_Se = 11.264f;
		break;

	case MATERIAL_RUDY:
		mat_Ka[0] = 0.1745f;
		mat_Kd[0] = 0.61424f;
		mat_Ks[0] = 0.727811f;
		mat_Ka[1] = 0.01175f;
		mat_Kd[1] = 0.04136f;
		mat_Ks[1] = 0.626959f;
		mat_Ka[2] = 0.01175f;
		mat_Kd[2] = 0.04136f;
		mat_Ks[2] = 0.626959f;
		mat_Ka[3] = 0.55f;
		mat_Kd[3] = 0.55f;
		mat_Ks[3] = 0.55f;
		mat_Se = 76.8f;
		break;

	case MATERIAL_TURQUOISE:
		mat_Ka[0] = 0.1f;
		mat_Kd[0] = 0.396f;
		mat_Ks[0] = 0.297254f;
		mat_Ka[1] = 0.18725f;
		mat_Kd[1] = 0.74151f;
		mat_Ks[1] = 0.30829f;
		mat_Ka[2] = 0.1745f;
		mat_Kd[2] = 0.69102f;
		mat_Ks[2] = 0.306678f;
		mat_Ka[3] = 0.8f;
		mat_Kd[3] = 0.8f;
		mat_Ks[3] = 0.8f;
		mat_Se = 12.8f;
		break;

	case MATERIAL_SKIN:
		mat_Ka[0] = 0.1f;
		mat_Ka[1] = 0.2f;
		mat_Ka[2] = 0.2f;
		mat_Kd[0] = 0.9f;
		mat_Kd[1] = 0.5f;
		mat_Kd[2] = 0.4f;
		mat_Ks[0] = 0.5f;
		mat_Ks[1] = 0.5f;
		mat_Ks[2] = 0.5f;
		mat_Se = 11.264f;
		break;


	case MATERIAL_INDIANRED:
		mat_Ka[0] = 0.2f;
		mat_Ka[1] = 0.2f;
		mat_Ka[2] = 0.2f;
		mat_Kd[0] = 1.0f;
		mat_Kd[1] = 0.42f;
		mat_Kd[2] = 0.42f;
		mat_Ks[0] = 0.5f;
		mat_Ks[1] = 0.5f;
		mat_Ks[2] = 0.5f;
		mat_Se = 15.0f;
		break;

	case MATERIAL_DODGEBLUE:
		mat_Ka[0] = 0.2f;
		mat_Ka[1] = 0.2f;
		mat_Ka[2] = 0.2f;
		mat_Kd[0] = 0.12f;
		mat_Kd[1] = 0.56f;
		mat_Kd[2] = 1.0f;
		mat_Ks[0] = 0.5f;
		mat_Ks[1] = 0.5f;
		mat_Ks[2] = 0.5f;
		mat_Se = 15.0f;
		break;

	case MATERIAL_DEFAULT:
	default:
		mat_Ka[0] = 0.3f;
		mat_Ka[1] = 0.3f;
		mat_Ka[2] = 0.3f;
		mat_Kd[0] = 1.0f;
		mat_Kd[1] = 1.0f;
		mat_Kd[2] = 1.0f;
		mat_Ks[0] = 0.7f;
		mat_Ks[1] = 0.7f;
		mat_Ks[2] = 0.7f;
		mat_Se = 64.0f;
		break;
		break;

	}

	glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_Kd);
	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_Ks);
	glMaterialfv(GL_FRONT, GL_AMBIENT, mat_Ka);
	glMaterialf(GL_FRONT, GL_SHININESS, mat_Se);

	mat_Ka[0] = 0.2f;
	mat_Ka[1] = 0.2f;
	mat_Ka[2] = 0.2f;
	mat_Kd[0] = 0.3f;  //backface color set here!!!
	mat_Kd[1] = 0.0f;
	mat_Kd[2] = 0.0f;
	mat_Ks[0] = 0.7f;
	mat_Ks[1] = 0.3f;
	mat_Ks[2] = 0.3f;
	mat_Se = 100.f;

	glMaterialfv(GL_BACK, GL_AMBIENT, mat_Ka);
	glMaterialfv(GL_BACK, GL_DIFFUSE, mat_Kd);
	glMaterialfv(GL_BACK, GL_SPECULAR, mat_Ks);
	glMaterialf(GL_BACK, GL_SHININESS, mat_Se);
}

void DisplayWindow::DoEnableLights()
{
	glEnable(GL_LIGHTING);

	if (lights_on[0])
		glEnable(GL_LIGHT0);

	if (lights_on[1])
		glEnable(GL_LIGHT1);

	if (lights_on[2])
		glEnable(GL_LIGHT2);

	if (lights_on[3])
		glEnable(GL_LIGHT3);
}

void DisplayWindow::DoDisableLights()
{
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHT1);
	glDisable(GL_LIGHT2);
	glDisable(GL_LIGHT3);
	glDisable(GL_LIGHTING);
}



//Some constant/static variables
const float DisplayWindow::sharpFaceColor[] = {0.0f, 0.4f, 0.9f};
const float DisplayWindow::sharpVerColor[] = {1.0f, 1.0f, 0.0f};
const float DisplayWindow::sharpEdgeColor[] = {1.0f, 1.0f, 0.0f};
const float DisplayWindow::surfaceColor[] = {1.0f, 1.0f, 0.6f};
const float DisplayWindow::normalColor[] = {1.0f,1.0f,1.0f};
const float DisplayWindow::edgeColor[] = {0.0f,0.0f,0.0f};
GLfloat DisplayWindow::R = 0.7f; 
GLfloat DisplayWindow::G = 0.7f; 
GLfloat DisplayWindow::B = 0.7f;
